@@ -302,6 +302,19 @@ sub load_image {
          my $magic = Gtk2::CV::magic $path;
          $type = $othertypes{$magic}
             if exists $othertypes{$magic};
+
+         if ($type =~ /^application\//) {
+            my $ffmpeg = qx{ffprobe -v quiet -show_format -of json -- \Q$path};
+
+            # ffmpeg doesn't properly encode json
+            $ffmpeg = eval { JSON::XS->new->latin1->decode ($ffmpeg) };
+
+            if ($ffmpeg->{format}{format_name} =~ /^image/) {
+               $type = "image/ffmpeg-$ffmpeg->{format}{format_name}";
+            } elsif (length $ffmpeg->{format}{format_name}) {
+               $type = "video/ffmpeg-$ffmpeg->{format}{format_name}";
+            }
+         }
       }
 
       if ((!$type or $type =~ /^application\/octet-stream/) && $path =~ /\.([^.]+)$/) {
@@ -312,6 +325,11 @@ sub load_image {
 
       $@ = "generic file display error";
 
+      if ($type eq "image/gif") {
+         # could be animation, very hard to detect, for the time being, force mpv
+         $type = "video/gif";
+      }
+
       if (!$type or $type =~ /^text\//) {
          $@ = "unrecognised file format";
          # should, but can't
@@ -501,7 +519,7 @@ sub load_image {
                   @mplayer_args = (
                      qw(
                         --no-input-terminal --no-input-default-bindings --no-input-cursor --input-conf=/dev/null
-                        --input-vo-keyboard=no
+                        --input-vo-keyboard=no -loop-file=inf
                      ),
                      @extra_mpv_args,
                      "--input-file=fd://" . fileno $sfh
@@ -510,7 +528,7 @@ sub load_image {
                   open STDOUT, ">/dev/null";
                   fcntl $sfh, Fcntl::F_SETFD (), 0;
                } else {
-                  @mplayer_args = qw(-nofs -slave -fixed-vo -nokeepaspect -input nodefault-bindings:conf=/dev/null -zoom);
+                  @mplayer_args = qw(-nofs -slave -fixed-vo -nokeepaspect -input nodefault-bindings:conf=/dev/null -zoom -loop 0);
                   open STDIN, "+<&" . fileno $sfh;
                   open STDOUT, ">&" . fileno $sfh;
                   # $sfh will autoclose
@@ -518,7 +536,7 @@ sub load_image {
 
                #open STDERR, ">/dev/null";
 
-               exec $Gtk2::CV::MPLAYER, @mplayer_args, qw(-quiet -loop 0),
+               exec $Gtk2::CV::MPLAYER, @mplayer_args, qw(-quiet),
                                @{ $rotate[$self->{load_rotate} / 90] },
                                -wid => $xid, $path;
                POSIX::_exit 0;
@@ -892,7 +910,11 @@ sub handle_key {
 
    my $volume = $Gtk2::CV::MPLAYER_IS_MPV ? "add ao-volume %s2" : "volume %s1";
 
-   if ($state * "control-mask") {
+   $state *= Gtk2::Accelerator->get_default_mod_mask;
+
+   $state -= ["shift-mask"];
+
+   if ($state == ["control-mask"]) {
       if ($key == $Gtk2::Gdk::Keysyms{p} || $key == $Gtk2::Gdk::Keysyms{P}) {
          new Gtk2::CV::PrintDialog
             pixbuf => $self->{subimage},
@@ -924,7 +946,7 @@ sub handle_key {
          return 0;
       }
 
-   } else {
+   } elsif ($state == []) {
       if ($key == $Gtk2::Gdk::Keysyms{less}) {
          $self->resize ($self->{dw} * 0.5, $self->{dh} * 0.5);
 
@@ -1018,7 +1040,7 @@ sub handle_key {
       } elsif ($self->{player_pid} && $key == $Gtk2::Gdk::Keysyms{o}) {
          print {$self->{mplayer_fh}} "no-osd cycle-values osd-level 2 3 0 1\n";
       } elsif ($self->{player_pid} && $key == $Gtk2::Gdk::Keysyms{p}) {
-         print {$self->{mplayer_fh}} "pause\n";
+         print {$self->{mplayer_fh}} "cycle pause\n";
       } elsif ($self->{player_pid} && $key == $Gtk2::Gdk::Keysyms{Escape}) {
          print {$self->{mplayer_fh}} "quit\n";
       } elsif ($self->{player_pid} && $key == $Gtk2::Gdk::Keysyms{9}) {
@@ -1032,6 +1054,8 @@ sub handle_key {
 
          return 0;
       }
+   } else {
+      return 0;
    }
 
    1
