@@ -32,8 +32,6 @@ sub INIT_INSTANCE {
       my $key = $_[1]->keyval;
       my $state = $_[1]->state;
 
-      my $ctrl = grep $_ eq "control-mask", @{$_[1]->state};
-
       if ($key == $Gtk2::Gdk::Keysyms{Up}) {
          return 0;
       } elsif ($key == $Gtk2::Gdk::Keysyms{Down}) {
@@ -183,6 +181,8 @@ sub new_schnauzer {
       my $key = $event->keyval;
       my $state = $event->state;
 
+      $state *= Gtk2::Accelerator->get_default_mod_mask;
+
       if (
          $state == ["control-mask"]
          && ($key == $Gtk2::Gdk::Keysyms{Up} || $key == $Gtk2::Gdk::Keysyms{Down})
